/*
 * Icons.cpp - Threaded shell icon retrieval.
 */

#define INC_OLE2
#define STRICT
#define COM_NO_WINDOWS_H

#include <windows.h>
#include <shellapi.h>
#include <process.h>
#include "icons.h"

static DWORD const s_idThread = GetCurrentThreadId();	// Main thread ID.
static HANDLE s_hBgThread = 0;							// Background thread handle.
static HANDLE s_hWake = 0;								// Event to wake the background thread.
static HANDLE s_hConfirmKilled = 0;						// Event to confirm background thread has ended.
static bool s_fKillThread = false;						// Causes background thread to exit when awoken.

static bool s_fInitializedCritSec = false;
static CRITICAL_SECTION s_csList;						// Guards access to the linked list.

/*
 * AsyncShellIcon.
 */

#ifdef DEBUG
static volatile long s_cTotal = 0;
static unsigned s_cEnqueued = 0;
static unsigned s_cDequeued = 0;
#endif

UINT const AsyncShellIcon::c_wmIconAvailable = RegisterWindowMessage( "WmAsyncIconAvailable" );

AsyncShellIcon* AsyncShellIcon::s_pHead = 0;
AsyncShellIcon* AsyncShellIcon::s_pTail = 0;

void
AsyncShellIcon::Uninitialize( bool const fWeakWait )
{
	if( s_hBgThread )
	{
		Assert( s_hWake );

		s_fKillThread = true;
		SetEvent( s_hWake );

		if( fWeakWait )
		{
			// Ideally we'd assert if the process is exiting and we reach
			// vsDllExit while s_hBgThread is true.  However in vsDllExit we
			// can't tell whether the process is exiting.
			WaitForSingleObject( s_hConfirmKilled, INFINITE );
			// Injecting a Sleep(0) here seems to perturb the timing (at least
			// on my machines) such that the background thread doesn't crash
			// even though we're about to unload the DLL out from under
			// running code.
			Sleep( 0 );
		}
		else
		{
#if 1
			WaitForSingleObject( s_hBgThread, INFINITE );
#else
			DWORD dwIndex;
			CoWaitForMultipleHandles( 0x07, INFINITE, 1, &s_hBgThread, &dwIndex );
#endif
		}

		CloseHandle( s_hBgThread );
		CloseHandle( s_hWake );
		CloseHandle( s_hConfirmKilled );
		s_hBgThread = 0;
		s_hWake = 0;
		s_hConfirmKilled = 0;
	}
}

#ifdef USE_CRT_THREADING
unsigned __stdcall
#else
DWORD WINAPI
#endif
AsyncShellIcon::ThreadProc( void* pv )
{
	Assert( s_hWake );
	Assert( s_hConfirmKilled );

	OleInitialize( 0 );	// So that SHGetFileInfo can be called for extracting icons.

	while( !s_fKillThread )
	{
		Trace( "AsyncShellIcon::threadproc waiting" );

		DWORD const dwWait = WaitForSingleObject( s_hWake, INFINITE );

		if( s_fKillThread )
			break;

		if( dwWait != WAIT_OBJECT_0 )
		{
			Panic();
			break;
		}

		Trace( "AsyncShellIcon::threadproc woke up" );

		while( s_pHead )
		{
			if( s_fKillThread )
				break;

			EnterCriticalSection( &s_csList );
			AsyncShellIcon* const p = s_pHead;
			if( p )
			{
				Assert( p->GetRefCount() == 2 );
				Assert( p->m_fQueued );
			}
			LeaveCriticalSection( &s_csList );

			if( !p )
				break;

			Assert( !p->IsComplete() );
			p->AsyncGetIcon();
		}

#ifdef DEBUG
		EnterCriticalSection( &s_csList );
		if( !s_pHead )
		{
			Assert( !s_pHead );
			Assert( !s_pTail );
			Assert( s_cEnqueued == s_cDequeued );
		}
		LeaveCriticalSection( &s_csList );
#endif
	}

	SetEvent( s_hConfirmKilled );

	return 0;
}

AsyncShellIcon::AsyncShellIcon()
	: m_hwndNotify( 0 )
	, m_fQueued( false )
	, m_fComplete( false )
#ifdef DEBUG
	, m_iIcon( -1 )
	, m_fZombie( false )
#endif
	, m_pNext( 0 )
	, m_pPrev( 0 )
{
	Assert( s_idThread == GetCurrentThreadId() );

	if( !s_fInitializedCritSec )
	{
		s_fInitializedCritSec = true;
		InitializeCriticalSection( &s_csList );
	}

	Assert( !m_hwndNotify );
	Assert( !m_fQueued );

	Assert( !m_pNext );
	Assert( !m_pPrev );

	Assert( !m_fComplete );

#ifdef DEBUG
	m_fZombie = false;
	InterlockedIncrement( &s_cTotal );
#endif
}

AsyncShellIcon::~AsyncShellIcon()
{
	Finish();

#ifdef DEBUG
	InterlockedDecrement( &s_cTotal );
#endif
}

void
AsyncShellIcon::Enqueue( HWND hwndNotify, char const* pszFile, ULONG ulShgfi = SHGFI_SMALLICON )
{
	Assert( hwndNotify );
	Assert( pszFile );
	Assert( !m_pNext );
	Assert( !m_pPrev );
	Assert( !m_sFile.Length() );
	Assert( !m_fComplete );
	Assert( !m_fZombie );

	// Record the details of the file for whom to retrieve an icon.

	m_hwndNotify = hwndNotify;
	m_sFile.Set( pszFile );
	m_ulShgfi = ulShgfi;

	while( m_sFile.Length() && m_sFile.Text()[m_sFile.Length() - 1] == '\\' )
		m_sFile.SetLength( m_sFile.Length() - 1 );

	// If the thread has already been killed then don't revive it.

	if( s_fKillThread )
	{
		PanicSz( "BUGBUG:  AsyncShellIcon background thread has already been killed; no further icon retrieval requests are allowed!" );
		return;
	}

	// Add the retrieval request to the async queue.

	EnterCriticalSection( &s_csList );
	AddRef();
	m_pPrev = s_pTail;
	if ( s_pTail )
		s_pTail->m_pNext = this;
	else
		s_pHead = this;
	s_pTail = this;
	m_fQueued = true;
#ifdef DEBUG
	s_cEnqueued++;
#endif
	LeaveCriticalSection( &s_csList );

	// Ensure the background thread.

	if ( !s_hBgThread )
	{
		Assert( !s_fKillThread );	// Once killed, no more requests should occur.

		if ( !s_hWake )
			s_hWake = CreateEvent( 0, false, false, 0 );
		if ( !s_hConfirmKilled )
			s_hConfirmKilled = CreateEvent( 0, false, false, 0 );
		s_fKillThread = false;

#ifdef USE_CRT_THREADING
		unsigned dwIndex;
		s_hBgThread = HANDLE( _beginthreadex( 0, 256*1024, AsyncShellIcon::ThreadProc, 0, 0, &dwIndex ) );
#else
		DWORD id;
		s_hBgThread = CreateThread( 0, 256*1024, AsyncShellIcon::ThreadProc, 0, 0, &id );
#endif
	}

	// Wake the background thread.

	if ( s_hBgThread )
		SetEvent( s_hWake );
}

void
AsyncShellIcon::Cancel()
{
	EnterCriticalSection( &s_csList );
	m_hwndNotify = 0;
	Finish();
	m_fComplete = true;
#ifdef DEBUG
	m_fZombie = true;
#endif
	LeaveCriticalSection( &s_csList );
}

int
AsyncShellIcon::RetrieveIconIndex()
{
	Assert( s_idThread == GetCurrentThreadId() );
	Assert( IsComplete() );
	Assert( !m_fZombie );
	Assert( !m_fQueued );

	return m_iIcon;
}

void
AsyncShellIcon::AsyncGetIcon()
{
	Assert( s_idThread != GetCurrentThreadId() );

	// Get the file icon.

	SHFILEINFO shfi;
	DWORD dwAttr = 0;
	ULONG ulFlags = SHGFI_SYSICONINDEX|m_ulShgfi;

	if( m_sFile.Equal( ".." ) ||
		( m_sFile.Length() >= 3 && !strcmp( m_sFile.Text() + m_sFile.Length() - 3, "\\.." ) ) )
	{
		dwAttr = FILE_ATTRIBUTE_DIRECTORY;
		ulFlags |= SHGFI_USEFILEATTRIBUTES;
	}

	if( SHGetFileInfo( m_sFile.Text(), dwAttr, &shfi, sizeof( shfi ), ulFlags ) )
		m_iIcon = shfi.iIcon;

	// Mark the retrieval request as finished and remove it from the queue.

	Finish();

	// Notify the caller's window.

	EnterCriticalSection( &s_csList );
	HWND const hwndNotify = m_hwndNotify;
	m_hwndNotify = 0;
	if ( hwndNotify )
		PostMessage( hwndNotify, c_wmIconAvailable, 0, 0 );
	LeaveCriticalSection( &s_csList );
}

void
AsyncShellIcon::Finish()
{
	EnterCriticalSection( &s_csList );
	bool const fQueued = m_fQueued;
	if ( fQueued )
	{
		if ( m_pPrev )
			m_pPrev->m_pNext = m_pNext;
		else
			s_pHead = m_pNext;
		if ( m_pNext )
			m_pNext->m_pPrev = m_pPrev;
		else
			s_pTail = m_pPrev;
		m_pNext = 0;
		m_pPrev = 0;
		m_fQueued = false;
#ifdef DEBUG
		s_cDequeued++;
#endif
	}

	m_sFile.Free();

	if ( fQueued )
		Release();
	LeaveCriticalSection( &s_csList );

	m_fComplete = true;
}

