/*
 * Str.h - Simplistic string class.
 */

#include <windows.h>
#include <stdio.h>
#include "std.h"
#include "str.h"
#ifndef USE_LFH_FOR_STR
#include "stdlib.h"
#endif

char const* const Str::s_pszEmpty = "";
#ifdef USE_LFH_FOR_STR
HANDLE Str::s_hHeap = 0;
#endif

void
Str::InitializeHeap()
{
#ifdef USE_LFH_FOR_STR
	Assert( !s_hHeap );

	// Allocate our own heap.  This is so we can enable LFH without any impact
	// to the default Windows heap in the SlickEdit process.

	s_hHeap = HeapCreate( 0, 0, 0 );

	// Enable low fragmentation heap.

	ULONG info = 2;
	HeapSetInformation( s_hHeap, HeapCompatibilityInformation, &info, sizeof( info ) );
#endif
}

void
Str::UninitializeHeap()
{
#ifdef USE_LFH_FOR_STR
	if( s_hHeap )
	{
		HeapDestroy( s_hHeap );
		s_hHeap = 0;
	}
#endif
}

inline char*
Str::MemReAlloc( char* pszExisting, size_t cchSize )
{
	void* pv;

#ifdef USE_LFH_FOR_STR
	if( pszExisting )
		pv = HeapReAlloc( s_hHeap, 0, pszExisting, cchSize );
	else
		pv = HeapAlloc( s_hHeap, 0, cchSize );
#else
	if( pszExisting )
		pv = realloc( pszExisting, cchSize );
	else
		pv = malloc( cchSize );
#endif

	return reinterpret_cast<char*>( pv );
}

inline void
Str::MemFree( char* pszFree )
{
#ifdef USE_LFH_FOR_STR
	HeapFree( s_hHeap, 0, pszFree );
#else
	free( pszFree );
#endif
}

Str::~Str()
{
	if( s_pszEmpty != m_psz )
		MemFree( m_psz );
}

void
Str::Clear()
{
	if( m_cchLen )
	{
		m_psz[0] = 0;
		m_cchLen = 0;
	}
}

void
Str::Free()
{
	if( s_pszEmpty != m_psz )
		MemFree( m_psz );

	Init();
}

void
Str::Init()
{
#ifdef USE_LFH_FOR_STR
	Assert( s_hHeap );
#endif

	m_psz = const_cast<char*>( s_pszEmpty );
	m_cchLen = 0;
	m_cchAlloc = 0;
}

bool
Str::Equal( Str const* s ) const
{
	Assert( s );

	if( Length() != s->Length() )
		return false;

	if( lstrcmp( Text(), s->Text() ) )
		return false;

	return true;
}

bool
Str::EqualI( Str const* s ) const
{
	Assert( s );

	if( Length() != s->Length() )
		return false;

	if( lstrcmpi( Text(), s->Text() ) )
		return false;

	return true;
}

void
Str::SetEnd( char const* pszEnd )
{
	SetLength( pszEnd - m_psz );
}

void
Str::SetLength( size_t cch )
{
	Assert( s_pszEmpty != m_psz );
	Assert( cch <= m_cchLen );

	m_cchLen = cch;
	m_psz[m_cchLen] = 0;
}

void
Str::Set( char const* psz, size_t cch )
{
	Clear();
	Append( psz, cch );
}

void
Str::Append( char ch )
{
	Assert( ch );

	memcpy( AllocInternal( 1 ), &ch, sizeof( ch ) );
	m_cchLen++;
	m_psz[m_cchLen] = 0;
}

void
Str::Append( char const* psz, size_t cch )
{
	if( psz )
	{
		if( int( cch ) < 0 )
			cch = lstrlen( psz );
		memcpy( AllocInternal( cch ), psz, cch );
		m_cchLen += cch;
		m_psz[m_cchLen] = 0;
	}
}

char*
Str::Alloc( size_t cch )
{
	m_cchLen += cch;
	return AllocInternal( 0 ) - cch;
}

char*
Str::AllocInternal( size_t cch )
{
	cch++;				// make room for null terminator

	if( m_cchLen + cch > m_cchAlloc )
	{
		// Alloc exact length the first time, but use geometric growth if
		// there is already a buffer allocated.

		size_t cchAlloc = m_cchLen + cch;

		if( m_cchAlloc > 31 )
			cchAlloc += ( cchAlloc / 3 ) * 2;
		else
			cchAlloc *= 2;

		// Allocate the new buffer.

		m_psz = MemReAlloc( ( s_pszEmpty == m_psz ) ? 0 : m_psz, cchAlloc  );
		m_cchAlloc = cchAlloc;
	}

	return m_psz + m_cchLen;
}

inline void
__HelperToEnsureBufferWillBeNullTerminated( void* const pv, size_t& cchMax, size_t const cbChar )
{
	// The "sn" variety of printf functions did not ensure the buffer is null
	// terminated.  To force the buffer to always be null terminated, we zero
	// out the entire buffer space, and then decrement the buffer size so
	// subsequent use will not overwrite the null at the end.

	ZeroMemory( pv, cchMax * cbChar );

	if( cchMax )
		cchMax--;
}

int
vsprintf_s( char* pszOut, size_t cchMax, char const* pszFormat, va_list args )
{
	__HelperToEnsureBufferWillBeNullTerminated( pszOut, cchMax, sizeof( *pszOut ) );

	return _vsnprintf( pszOut, cchMax, pszFormat, args );
}

void
Str::PrintfV( char const* pszFormat, va_list args )
{
	size_t cch = size_t( -1 );
	size_t cchAlloc = 100;	// start with 100

	while( int( cch ) < 0 )
	{
		// Subtract 1 from the max character count to ensure room for null
		// terminator; see MSDN for idiosyncracy of the snprintf family of
		// functions.

		cch = vsprintf_s( AllocInternal( cchAlloc ), cchAlloc - 1, pszFormat, args );
		cchAlloc *= 2;
	}

	m_cchLen += cch;
}

void
Str::Printf( char const* pszFormat, ... )
{
	va_list args;
	va_start( args, pszFormat );

	PrintfV( pszFormat, args );

	va_end( args );
}

#if 0

template <class T>
void
Str<T>::Printf( UINT idsFormat, ... )
{
	va_list args;
	va_start( args, idsFormat );

	TCHAR szFormat[100];
	LPCTSTR pszFormat = szFormat;
	StrT sFormat;

	if( LoadStringU( idsFormat, szFormat, DimensionOf( szFormat ) ) + 1 == DimensionOf( szFormat ) )
	{
	    sFormat.LoadString( idsFormat );
	    pszFormat = sFormat.Text();
	}

	PrintfV( pszFormat, args );

	va_end( args );
}

template <class T>
size_t
Str<T>::LoadString( UINT ids )
{
	for( ;; )
	{
	    Clear();

	    if( m_cchAlloc )
		m_cchLen = LoadStringU( ids, m_psz, m_cchAlloc );

	    // When loading a resource string in Windows, the only way to
	    // detect that the buffer may not have been large enough is to
	    // check if the string exactly fills the buffer.

	    if( m_cchLen + 1 < m_cchAlloc )
		break;

	    AllocInternal( Max( size_t( 31 ), m_cchAlloc ) );
	}

	return m_cchLen;
}

#endif

size_t
Str::GetWindowText( HWND hwnd )
{
	int const cch = ::GetWindowTextLength( hwnd );

	Clear();

	if( cch )
	{
	    AllocInternal( cch );
	    m_cchLen = ::GetWindowText( hwnd, m_psz, m_cchAlloc );
	}

	return m_cchLen;
}

void
Str::TrimLeft()
{
	if( Length() )
	{
		char const* psz = Text();

		while( *psz && isspace( *psz ) )
			psz++;

		if( psz > Text() )
		{
			m_cchLen -= int( psz - m_psz );
			memmove( m_psz, psz, ( m_cchLen + 1 ) );
		}
	}
}

void
Str::TrimRight()
{
	if( Length() )
	{
		char const* pszLast = Text() + Length() - 1;

		while( pszLast >= Text() && isspace( *pszLast ) )
			pszLast--;

		SetEnd( pszLast + 1 );
	}
}

void
Str::Trim()
{
	TrimRight();
	TrimLeft();
}

#if 0

template <class T>
bool
Str<T>::GetToken( Str<T> &out )
{
	// Clear out token string.

	out.Clear();

	if( s_pszEmpty == m_psz )
	    return 0;

	// Parse the string for a comma separated token.  Newline and EOS are
	// also considered separators.

	bool fQuote = false;
	T const* psz = Text();

	for( ; *psz; psz++ )
	{
	    if( '\"' == *psz )
	    {
		if( !fQuote && !out.Length() )
		{
		    fQuote = true;
		    continue;
		}
		else if( fQuote )
		{
		    fQuote = false;

		    psz++;

		    if( !*psz )
		    {
			// End of string.
			break;
		    }

		    if( ',' == *psz || '\n' == *psz )
		    {
			// End of token.
			psz++;
			break;
		    }

		    // Two double quotes in a row are interpreted as one
		    // double quote and do not leave quote mode.

		    if( '\"' == *psz )
		    {
			fQuote = true;
			out.Append( psz, 1 );
			continue;
		    }

		    // For poorly terminated tokens, skip up to the next
		    // delimeter.

		    while( *psz )
		    {
			if( '\n' == *psz || ',' == *psz )
			{
			    psz++;
			    break;
			}
		    }

		    break;
		}
	    }

	    if( !fQuote && ( '\n' == *psz || ',' == *psz ) )
	    {
		psz++;
		break;
	    }

	    out.Append( psz, 1 );
	}

	// Chop off the parsed token from the beginning of the string.  This
	// is where an iterator like StrRef would come in handy.

	m_cchLen -= int( psz - m_psz );
	memmove( m_psz, psz, ( m_cchLen + 1 ) );

	return !!out.Length();
}

#endif

#if 0

template <class T>
void
Str<T>::Attach( T* psz, size_t cchLen, size_t cchAlloc )
{
	if( s_pszEmpty != m_psz )
	    free( m_psz );

	if( psz )
	{
	    Assert( cchLen );
	    Assert( cchAlloc );
	    Assert( cchAlloc > cchLen );

	    m_cchLen = cchLen;
	    m_cchAlloc = cchAlloc;

	    m_psz = psz;
	    m_psz[cchLen] = 0;
	}
	else
	{
	    m_cchLen = 0;
	    m_cchAlloc = 0;
	    m_psz = const_cast<T*>( s_pszEmpty );
	}
}

template <class T>
void
Str<T>::Swap( Str<T>& other )
{
	T* const psz = m_psz;
	m_psz = other.m_psz;
	other.m_psz = psz;

	size_t const cchLen = m_cchLen;
	m_cchLen = other.m_cchLen;
	other.m_cchLen = cchLen;

	size_t const cchAlloc = m_cchAlloc;
	m_cchAlloc = other.m_cchAlloc;
	other.m_cchAlloc = cchAlloc;
}

#endif

